/*!@file
** @brief   The entry point of the program.
** @remarks This file is the single compilation unit of the program.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2014-2020 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#include "stdafx.h"
#include "resource.h"
#include "about.h"
#include "builder.h"
#include "clipboard.h"

#include <shlobj.h>

/*!
** @brief   The global instance of CComModule.
** @remarks -
*/
CComModule _Module;

namespace
{
	/*!
	** @brief   Checks to see if the path exists.
	** @param   [in] pszPath the path to be verified as valid for writing.
	** @param   [in] hwnd the parent window to be used for any user interface windows.
	** @param   [in] dwFlags determines behavior options.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL PrepareForWrite(LPCTSTR pszPath, HWND hwnd = NULL, DWORD dwFlags = SHPPFW_DEFAULT)
	{
		CString strFullPath;
		::GetFullPathName(pszPath, MAX_PATH, strFullPath.GetBuffer(MAX_PATH), NULL);
		strFullPath.ReleaseBuffer();
		ATLASSERT(!strFullPath.IsEmpty());

		HRESULT hResult = ::SHPathPrepareForWrite(hwnd, NULL, strFullPath, dwFlags);

		if (SUCCEEDED(hResult))
		{
			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @brief   Deletes an existing directory recursively.
	** @param   [in] pszPath the path of the directory to be removed.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL RemoveDirectoryRecursively(LPCTSTR pszPath)
	{
		CString strFullPath;
		::GetFullPathName(pszPath, MAX_PATH, strFullPath.GetBuffer(MAX_PATH), NULL);
		strFullPath.ReleaseBuffer();
		ATLASSERT(!strFullPath.IsEmpty());

		if (ATLPath::IsDirectory(strFullPath))
		{
			SHFILEOPSTRUCT op = {};

			op.wFunc = FO_DELETE, op.pFrom = strFullPath, op.fFlags = FOF_NO_UI;

			if (::SHFileOperation(&op) == 0)
			{
				return TRUE;
			}
		}
		return FALSE;
	}
};

/*!
** @brief   Implements the entry point for the application.
** @param   [in] hInstance a handle to the current instance of the application.
** @param   [in] hPrevInstance a legacy from 16-bit Windows, now always zero.
** @param   [in] lpCmdLine the command-line for the application.
** @param   [in] nCmdShow specifies how the window is to be shown.
** @returns A standard HRESULT value as the exit value.
** @remarks -
*/
extern "C" int WINAPI _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	UNREFERENCED_PARAMETER(hInstance);
	UNREFERENCED_PARAMETER(hPrevInstance);
	UNREFERENCED_PARAMETER(lpCmdLine);
	UNREFERENCED_PARAMETER(nCmdShow);

	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);

	HRESULT hResult = S_OK;

	hResult = ::CoInitialize(NULL);
	ATLASSERT(SUCCEEDED(hResult));
	::DefWindowProc(NULL, WM_NULL, 0, 0L);
	ATLASSERT(!CAtlBaseModule::m_bInitFailed);
	hResult = _Module.Init(NULL, hInstance);
	ATLASSERT(SUCCEEDED(hResult));

	#if 1 // Setup
	{
		CPath path = TEXT(".\\target\\junk\\FileName.Ext");

		ATLVERIFY(PrepareForWrite(path, NULL, SHPPFW_DEFAULT | SHPPFW_IGNOREFILENAME));

		CAtlFile file;
		hResult = file.Create(path, GENERIC_WRITE, 0, CREATE_ALWAYS);
		ATLASSERT(SUCCEEDED(hResult));
		hResult = file.SetSize(8191);
		ATLASSERT(SUCCEEDED(hResult));

		FILETIME ftCurrentTime;
		::GetSystemTimeAsFileTime(&ftCurrentTime);
		ULARGE_INTEGER nCurrentTime = {
			ftCurrentTime.dwLowDateTime,
			ftCurrentTime.dwHighDateTime,
		};
		nCurrentTime.QuadPart = nCurrentTime.QuadPart / 600000000 * 600000000;

		nCurrentTime.QuadPart += 10000000; // HH:mm:01
		FILETIME ftCreationTime = {
			nCurrentTime.LowPart,
			nCurrentTime.HighPart,
		};
		nCurrentTime.QuadPart += 10000000; // HH:mm:02
		FILETIME ftLastAccessTime = {
			nCurrentTime.LowPart,
			nCurrentTime.HighPart,
		};
		nCurrentTime.QuadPart += 10000000; // HH:mm:03
		FILETIME ftLastWriteTime = {
			nCurrentTime.LowPart,
			nCurrentTime.HighPart,
		};
		ATLVERIFY(::SetFileTime(file,
			&ftCreationTime,
			&ftLastAccessTime,
			&ftLastWriteTime
		));
		file.Close();
	}
	#endif

	#if 1 // Test 1: CClipboard::SetText
	{
		const wchar_t* pszExpected = L"Hello, world!";

		ATLVERIFY(CClipboard::SetText(pszExpected));

		ATLVERIFY(::OpenClipboard(NULL));

		int nAvailableFormats = 0;

		for (UINT uFormat = ::EnumClipboardFormats(0); uFormat; uFormat = ::EnumClipboardFormats(uFormat))
		{
			switch (uFormat)
			{
			case CF_LOCALE:
			case CF_OEMTEXT:
			case CF_TEXT:
			case CF_UNICODETEXT:

				ATLASSERT(::IsClipboardFormatAvailable(uFormat));
				break;

			default:

				ATLASSERT(FALSE);
				break;
			}
			++nAvailableFormats;
		}
		ATLASSERT(nAvailableFormats == 4);
		ATLASSERT(::CountClipboardFormats() == nAvailableFormats);

		HGLOBAL hGlobal = ::GetClipboardData(CF_UNICODETEXT);
		ATLASSERT(hGlobal);
		wchar_t* pszGlobal = static_cast<wchar_t*>(::GlobalLock(hGlobal));
		ATLASSERT(pszGlobal);
		ATLTRACE(TEXT("\"%ls\"\n"), pszGlobal);
		ATLASSERT(wcscmp(pszGlobal, pszExpected) == 0);
		if (::GlobalUnlock(hGlobal) == 0)
			ATLASSERT(::GetLastError() == NO_ERROR);

		ATLVERIFY(::CloseClipboard());
	}
	#endif

	#if 1 // Test 2: CTextBuilder::GetText
	{
		CString strText;
		DWORD dwLastError = ERROR_SUCCESS;
		int argc = 0;
		TCHAR *argv[] = {
			__targv[0],
			TEXT("arg1"),
			TEXT("arg2"),
			TEXT("arg3"),
		};

		#if 1 // empty
		{
			strText = TEXT("foobar"); // to be replaced by a new text
			ATLVERIFY(CTextBuilder::GetText(strText, argc = 1, argv));
			ATLASSERT(strText.IsEmpty());
			DBG_UNREFERENCED_LOCAL_VARIABLE(dwLastError);
		}
		#endif

		#if 1 // single-line text
		{
			strText = TEXT("foobar"); // to be replaced by a new text
			ATLVERIFY(CTextBuilder::GetText(strText, argc = 2, argv));
			ATLASSERT(strText == TEXT("arg1")); // no trailing newline
			DBG_UNREFERENCED_LOCAL_VARIABLE(dwLastError);
		}
		#endif

		#if 1 // multiline text
		{
			strText = TEXT("foobar"); // to be replaced by a new text
			ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
			ATLASSERT(strText == TEXT("arg1\r\narg2\r\n"));
			DBG_UNREFERENCED_LOCAL_VARIABLE(dwLastError);
		}
		#endif

		#if 1 // unknown option
		{
			CString strPreserved = strText = TEXT("foobar");

			argv[2] = TEXT("-unknown");

			::SetLastError(ERROR_SUCCESS);
			ATLVERIFY(!CTextBuilder::GetText(strText, argc = 4, argv));
			dwLastError = ::GetLastError();
			ATLASSERT(HRESULT(dwLastError) == E_INVALIDARG);
			ATLASSERT(strText == strPreserved);

			argv[2] = TEXT("/unknown");

			::SetLastError(ERROR_SUCCESS);
			ATLVERIFY(!CTextBuilder::GetText(strText, argc = 4, argv));
			dwLastError = ::GetLastError();
			ATLASSERT(HRESULT(dwLastError) == E_INVALIDARG);
			ATLASSERT(strText == strPreserved);
		}
		#endif
	}
	#endif

	#if 1 // Test 3: Command-line option /text
	{
		CString strText;
		DWORD dwLastError = ERROR_SUCCESS;
		int argc = 0;
		TCHAR *argv[] = {
			__targv[0],
			TEXT("/text"),
			TEXT("arg1"),
			TEXT("arg2"),
			TEXT("arg3"),
		};

		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == TEXT("arg1")); // no trailing newline
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 4, argv));
		ATLASSERT(strText == TEXT("arg1\r\narg2\r\n"));

		argv[3] = TEXT("-unknown");

		ATLVERIFY(CTextBuilder::GetText(strText, argc = 5, argv));
		ATLASSERT(strText == TEXT("arg1\r\n-unknown\r\narg3\r\n"));
		DBG_UNREFERENCED_LOCAL_VARIABLE(dwLastError);

		argv[3] = TEXT("/unknown");

		ATLVERIFY(CTextBuilder::GetText(strText, argc = 5, argv));
		ATLASSERT(strText == TEXT("arg1\r\n/unknown\r\narg3\r\n"));
		DBG_UNREFERENCED_LOCAL_VARIABLE(dwLastError);
	}
	#endif

	#if 1 // Test 4: Command-line option /path
	{
		CString strText;
		DWORD dwLastError = ERROR_SUCCESS;
		int argc = 0;
		TCHAR *argv[] = {
			__targv[0],
			TEXT("/path:*"),
			TEXT(".\\FileName.Ext"), // the file need not exist.
		};

		CString strPath;
		::GetCurrentDirectory(MAX_PATH, strPath.GetBuffer(MAX_PATH));
		strPath.ReleaseBuffer();
		ATLASSERT(!ATLPath::IsRelative(strPath));

		argv[1] = TEXT("/path:fullpath");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == strPath + TEXT("\\FileName.Ext"));
		argv[1] = TEXT("/path:filename");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == TEXT("FileName.Ext"));
		argv[1] = TEXT("/path:extension");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == TEXT(".Ext"));
		argv[1] = TEXT("/path:basename");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == TEXT("FileName"));
		argv[1] = TEXT("/path:parent");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == strPath);

		argv[1] = TEXT("/path:unknown");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == TYPE_E_ELEMENTNOTFOUND);

		argv[1] = TEXT("/path:");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == TYPE_E_ELEMENTNOTFOUND);

		argv[2] = TEXT("\\\\server\\share\\file"); // UNC

		argv[1] = TEXT("/path:filename");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == TEXT("file"));
		argv[1] = TEXT("/path:parent");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == TEXT("\\\\server\\share"));
	}
	#endif

	#if 1 // Test 5: Command-line option /prop
	{
		CString strText;
		DWORD dwLastError = ERROR_SUCCESS;
		int argc = 0;
		TCHAR *argv[] = {
			__targv[0],
			TEXT("/prop:*"),
			TEXT(".\\target\\junk\\FileName.Ext"),
		};

		CTextBuilder::s_pdfFlags = PDFF_DEFAULT;

		argv[1] = TEXT("/prop:System.FileName");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == TEXT("FileName.Ext"));
		argv[1] = TEXT("/prop:System.Size");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(strText == TEXT("8191"));

		argv[1] = TEXT("/prop:unknown");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == TYPE_E_ELEMENTNOTFOUND);

		argv[1] = TEXT("/prop:");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == TYPE_E_ELEMENTNOTFOUND);

		argv[2] = TEXT(".\\target\\junk\\unknown");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND));
	}
	#endif

	#if 1 // Test 6: Command-line option /display
	{
		CString strText;
		DWORD dwLastError = ERROR_SUCCESS;
		int argc = 0;
		TCHAR *argv[] = {
			__targv[0],
			TEXT("/display"),
			TEXT("/prop:*"),
			TEXT(".\\target\\junk\\FileName.Ext"),
		};

		CTextBuilder::s_pdfFlags = PDFF_DEFAULT;

		argv[2] = TEXT("/prop:System.FileName");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 4, argv));
		ATLASSERT(strText == TEXT("FileName.Ext"));
		argv[2] = TEXT("/prop:System.Size");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 4, argv));
		ATLASSERT(strText == TEXT("7.99 KB"));

		DBG_UNREFERENCED_LOCAL_VARIABLE(dwLastError);
	}
	#endif

	#if 1 // Test 7: Command-line option /date
	{
		CString strText;
		DWORD dwLastError = ERROR_SUCCESS;
		int argc = 0;
		TCHAR *argv[] = {
			__targv[0],
			TEXT("/date:*"),
			TEXT("/prop:System.Null"),
		};

		PROPDESC_FORMAT_FLAGS& pdfFlags = CTextBuilder::s_pdfFlags;

		pdfFlags = PDFF_SHORTTIME; // to be preserved

		argv[1] = TEXT("/date:short");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(pdfFlags == (PDFF_SHORTTIME | PDFF_SHORTDATE));
		argv[1] = TEXT("/date:long");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(pdfFlags == (PDFF_SHORTTIME | PDFF_LONGDATE));
		argv[1] = TEXT("/date:hide");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(pdfFlags == (PDFF_SHORTTIME | PDFF_HIDEDATE));

		argv[1] = TEXT("/date:unknown");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == E_INVALIDARG);

		argv[1] = TEXT("/date:");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == E_INVALIDARG);

		pdfFlags = PDFF_DEFAULT;
	}
	#endif

	#if 1 // Test 8: Command-line option /time
	{
		CString strText;
		DWORD dwLastError = ERROR_SUCCESS;
		int argc = 0;
		TCHAR *argv[] = {
			__targv[0],
			TEXT("/time:*"),
			TEXT("/prop:System.Null"),
		};

		PROPDESC_FORMAT_FLAGS& pdfFlags = CTextBuilder::s_pdfFlags;

		pdfFlags = PDFF_SHORTDATE; // to be preserved

		argv[1] = TEXT("/time:short");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(pdfFlags == (PDFF_SHORTDATE | PDFF_SHORTTIME));
		argv[1] = TEXT("/time:long");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(pdfFlags == (PDFF_SHORTDATE | PDFF_LONGTIME));
		argv[1] = TEXT("/time:hide");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLASSERT(pdfFlags == (PDFF_SHORTDATE | PDFF_HIDETIME));

		argv[1] = TEXT("/time:unknown");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == E_INVALIDARG);

		argv[1] = TEXT("/time:");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == E_INVALIDARG);

		pdfFlags = PDFF_DEFAULT;
	}
	#endif

	#if 1 // Test 9: Command-line option /stat
	{
		CString strText;
		DWORD dwLastError = ERROR_SUCCESS;
		int argc = 0;
		TCHAR *argv[] = {
			__targv[0],
			TEXT("/stat:*"),
			TEXT(".\\target\\junk\\FileName.Ext"),
		};

		COleDateTime varDateTime;

		argv[1] = TEXT("/stat:atime");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLVERIFY(varDateTime.ParseDateTime(strText));
		ATLASSERT(varDateTime.GetSecond() == 2);
		argv[1] = TEXT("/stat:mtime");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLVERIFY(varDateTime.ParseDateTime(strText));
		ATLASSERT(varDateTime.GetSecond() == 3);
		argv[1] = TEXT("/stat:ctime");
		ATLVERIFY(CTextBuilder::GetText(strText, argc = 3, argv));
		ATLVERIFY(varDateTime.ParseDateTime(strText));
		ATLASSERT(varDateTime.GetSecond() == 1);

		argv[1] = TEXT("/stat:unknown");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == TYPE_E_ELEMENTNOTFOUND);

		argv[1] = TEXT("/stat:");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == TYPE_E_ELEMENTNOTFOUND);

		argv[2] = TEXT(".\\target\\junk\\unknown");

		::SetLastError(ERROR_SUCCESS);
		ATLVERIFY(!CTextBuilder::GetText(strText, argc = 3, argv));
		dwLastError = ::GetLastError();
		ATLASSERT(HRESULT(dwLastError) == HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND));
	}
	#endif

	#if 1 // Teardown
	{
		RemoveDirectoryRecursively(TEXT(".\\target\\junk"));
	}
	#endif

	_Module.Term();
	::CoUninitialize();
	return int(hResult);
}
